/*-*- c++ -*-********************************************************
 * wxComposeView.cc : a wxWindows look at a message                 *
 *                                                                  *
 * (C) 1998-2000 by Karsten Ballder (Ballueder@gmx.net)            *
 *                                                                  *
 * $Id: wxComposeView.cpp,v 1.196 2001/03/13 19:34:43 vadz Exp $   *
 *******************************************************************/

#ifdef __GNUG__
#pragma  implementation "wxComposeView.h"
#endif

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#include "Mpch.h"
#include "Mcommon.h"

#ifndef USE_PCH
#  include "strutil.h"
#  include "sysutil.h"

#  include "PathFinder.h"
#  include "Profile.h"
#  include "MHelp.h"
#  include "MFrame.h"

#  include "MApplication.h"
#  include "gui/wxMApp.h"

#  include <wx/process.h>

#  include <ctype.h>          // for isspace()
#endif

#include <wx/textctrl.h>
#include <wx/confbase.h>      // for wxExpandEnvVars()
#include <wx/ffile.h>
#include <wx/textfile.h>
#include <wx/process.h>
#include <wx/mimetype.h>
#include <wx/tokenzr.h>
#include "wx/persctrl.h"

#include "Mdefaults.h"
#include "Mpers.h"

#include "FolderView.h"
#include "MailFolder.h"
#include "Message.h"
#include "SendMessage.h"

#include "MDialogs.h"
#include "HeadersDialogs.h"

#include "gui/wxIconManager.h"

#include "gui/wxllist.h"
#include "gui/wxlwindow.h"
#include "gui/wxlparser.h"
#include "gui/wxIdentityCombo.h"
#include "gui/wxOptionsDlg.h"
#include "gui/wxComposeView.h"

#include "adb/AdbEntry.h"
#include "adb/AdbManager.h"

#include "MessageTemplate.h"
#include "TemplateDialog.h"
#include "MModule.h"
#include "modules/Calendar.h"

// incredible, but true: cclient headers #define the symbol write...
#undef write

// ----------------------------------------------------------------------------
// constants
// ----------------------------------------------------------------------------

// control ids
enum
{
   IDB_EXPAND = 100
};

// the default message title
#define COMPOSER_TITLE (_("Message Composition"))

#define CANONIC_ADDRESS_SEPARATOR   ", "

// code here was written with assumption that x and y margins are the same
#define LAYOUT_MARGIN LAYOUT_X_MARGIN

// ----------------------------------------------------------------------------
// private functions
// ----------------------------------------------------------------------------

// return TRUE if the field should have address expansion
static inline bool IsAddressField(size_t field)
{
   return field != wxComposeView::Field_Subject;
}

static wxString GetMimeTypeFromFilename(const wxString& filename)
{
   wxString strExt;
   wxSplitPath(filename, NULL, NULL, &strExt);

   wxString strMimeType;
   wxMimeTypesManager& mimeManager = mApplication->GetMimeManager();
   wxFileType *fileType = mimeManager.GetFileTypeFromExtension(strExt);
   if ( (fileType == NULL) || !fileType->GetMimeType(&strMimeType) )
   {
      // can't find MIME type from file extension, set some default one
      strMimeType = "APPLICATION/OCTET-STREAM";
   }

   delete fileType;  // may be NULL, ok

   return strMimeType;
}

// ----------------------------------------------------------------------------
// private classes
// ----------------------------------------------------------------------------

class MimeContent : public wxLayoutObject::UserData
{
public:
   // constants
   enum MimeContentType
   {
      MIMECONTENT_NONE,
      MIMECONTENT_FILE,
      MIMECONTENT_DATA
   };

   // ctor & dtor
   MimeContent()
      {
         m_Type = MIMECONTENT_NONE;
      }
   // initialize
   void SetMimeType(const String& mimeType);
   void SetData(char *data, size_t length,
                const char *filename = NULL); // we'll delete data!
   void SetFile(const String& filename);

   // accessors
   MimeContentType GetType() const { return m_Type; }

   const String& GetMimeType() const { return m_MimeType; }
   Message::ContentType GetMimeCategory() const { return m_NumericMimeType; }

   const String& GetFileName() const
      { return m_FileName; }

   const char *GetData() const
      { ASSERT( m_Type == MIMECONTENT_DATA ); return m_Data; }
   size_t GetSize() const
      { ASSERT( m_Type == MIMECONTENT_DATA ); return m_Length; }

protected:
   ~MimeContent()
      {
         if ( m_Type == MIMECONTENT_DATA )
            delete [] m_Data;
      }


private:
   MimeContentType m_Type;

   char     *m_Data;
   size_t    m_Length;
   String    m_FileName;

   Message::ContentType m_NumericMimeType;
   String               m_MimeType;
};

// specialized text control which processes TABs to expand the text it
// contains and also notifies parent (i.e. wxComposeView) when it is
// modified.
//
// NB: use <Enter> to change the current control when wxAddressTextCtrl
//     has focus (TAB won't work!)
class wxAddressTextCtrl : public wxTextCtrl
{
public:
   // ctor
   wxAddressTextCtrl(wxComposeView *composeView,
                     wxComposeView::AddressField id,
                     wxWindow *parent)
      : wxTextCtrl(parent, -1, "",
                   wxDefaultPosition, wxDefaultSize,
                   wxTE_PROCESS_ENTER | wxTE_PROCESS_TAB)
   {
      m_composeView = composeView;
      m_id = id;
      m_lookupMode = (READ_CONFIG(m_composeView->GetProfile(),
                                  MP_ADB_SUBSTRINGEXPANSION) )
                     ? AdbLookup_Substring
                     : AdbLookup_StartsWith;
   }

   // expand the text in the control using the address book(s): returns FALSE
   // if no expansion took place
   bool DoExpand();

   // callbacks
   void OnFocusSet(wxFocusEvent& event);
   void OnChar(wxKeyEvent& event);
   void OnEnter(wxCommandEvent& event);

private:
   wxComposeView              *m_composeView;
   wxComposeView::AddressField m_id;

   int  m_lookupMode;

   DECLARE_EVENT_TABLE()
};

// a slightly different wxLayoutWindow: it calls back the composer the first
// time it gets focus which allows us to launch an external editor then (if
// configured to do so, of course)
class wxComposerLayoutWindow : public wxLayoutWindow
{
public:
   wxComposerLayoutWindow(wxComposeView *composer, wxWindow *parent)
      : wxLayoutWindow(parent)
   {
      m_composer = composer;
      m_firstTime = TRUE;
   }

protected:
   void OnFocus(wxFocusEvent& event)
   {
      if ( m_firstTime )
      {
         m_firstTime = FALSE;

         m_composer->OnFirstTimeFocus();
      }

      event.Skip();
   }

private:
   wxComposeView *m_composer;
   bool m_firstTime;

   DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// the classes which are used together with compose view - a derivation of
// variable expander and expansion sink
// ----------------------------------------------------------------------------

// this struct and array are used by ExpansionSink only - they contain the
// information about the attachments
struct AttachmentInfo
{
   AttachmentInfo(const void *d,
                  size_t l,
                  const String& m,
                  const String& f) : mimetype(m), filename(f)
      { data = d; len = l; }

   const void *data;
   size_t      len;
   String      mimetype,
               filename;
};

WX_DECLARE_OBJARRAY(AttachmentInfo, ArrayAttachmentInfo);

class ExpansionSink : public MessageTemplateSink
{
public:
   // ctor
   ExpansionSink() { m_hasCursorPosition = FALSE; m_x = m_y = 0; }

   // called after successful parsing of the template to insert the resulting
   // text into the compose view
   void InsertTextInto(wxComposeView& cv) const;

   // implement base class pure virtual function
   virtual bool Output(const String& text);

   // TODO the functions below should be in the base class (as pure virtuals)
   //      somehow, not here

   // called by VarExpander to remember the current position as the initial
   // cursor position
   void RememberCursorPosition() { m_hasCursorPosition = TRUE; }

   // called by VarExpander to insert an attachment
   void InsertAttachment(void *data, size_t len,
                         const String& mimetype,
                         const String& filename);

private:
   // as soon as m_hasCursorPosition becomes TRUE we stop to update the current
   // cursor position which we normally keep track of in m_x and m_y - so that
   // we'll have the position of the cursor when RememberCursorPosition() in
   // them at the end
   bool m_hasCursorPosition;
   int  m_x, m_y;

   // before each m_attachments[n] there is text from m_texts[n] - and after
   // the last attachment there is m_text
   wxString m_text;
   wxArrayString m_texts;
   ArrayAttachmentInfo m_attachments;
};

// VarExpander is the implementation of MessageTemplateVarExpander used here.
class VarExpander : public MessageTemplateVarExpander
{
public:
   // all categories
   enum Category
   {
      Category_Misc,       // empty category - misc variables
      Category_File,       // insert file
      Category_Attach,     // attach file
      Category_Command,    // execute external command
#ifdef USE_PYTHON
      Category_Python,     // execute Python script
#endif // USE_PYTHON
      Category_Message,    // the headers of the message being written
      Category_Original,   // variables pertaining to the original message
      Category_Invalid,    // unknown/invalid category
      Category_Max = Category_Invalid
   };

   // the variables in "misc" category
   enum Variable
   {
      Var_Date,            // insert the date in the default format
      Var_Cursor,          // position the cursor here after template expansion
      Var_To,              // the recepient name
      Var_Subject,         // the message subject (without any "Re"s)

      // all entries from here only apply to the reply/forward/followup
      // templates because they work with the original message
      Var_Quote,           // quote the original text
      Var_Quote822,        // include the original msg as a RFC822 attachment
      Var_Sender,          // the fullname of the sender

      Var_Invalid,
      Var_Max = Var_Invalid
   };

   // the variables in "message" category
   //
   // NB: the values should be identical to wxComposeView::AddressField enum
   //     (or the code in ExpandMessage should be changed)
   enum MessageHeader
   {
      MessageHeader_To,
      MessageHeader_Subject,
      MessageHeader_Cc,
      MessageHeader_Bcc,
      MessageHeader_LastControl = MessageHeader_Bcc,
      MessageHeader_FirstName,
      MessageHeader_LastName,
      MessageHeader_Invalid,
      MessageHeader_Max = MessageHeader_Invalid
   };

   // the variables from "original" category which map to headers
   enum OriginalHeader
   {
      OriginalHeader_Date,
      OriginalHeader_From,
      OriginalHeader_Subject,
      OriginalHeader_PersonalName,
      OriginalHeader_FirstName,
      OriginalHeader_LastName,
      OriginalHeader_To,
      OriginalHeader_ReplyTo,
      OriginalHeader_Newsgroups,
      OriginalHeader_Invalid,
      OriginalHeader_Max = OriginalHeader_Invalid
   };

   // get the category from the category string (may return Category_Invalid)
   static Category GetCategory(const String& category)
   {
      return (Category)FindStringInArray(ms_templateVarCategories,
                                         Category_Max, category);
   }

   // get the variable id from the string (works for misc variables only, will
   // return Var_Invalid if variable is unknown)
   static Variable GetVariable(const String& variable)
   {
      return (Variable)FindStringInArray(ms_templateVarNames,
                                         Var_Max, variable);
   }

   // get the header corresponding to the variable of "message" category
   static MessageHeader GetMessageHeader(const String& variable)
   {
      return (MessageHeader)FindStringInArray(ms_templateMessageVars,
                                              MessageHeader_Max, variable);
   }

   // get the header corresponding to the variable of "original" category
   // (will return OriginalHeader_Invalid if there is no corresponding header -
   // note that this doesn't mean that the variable is invalid because, for
   // example, "quote" doesn't correspond to any header, yet $(original:quote)
   // is perfectly valid)
   static OriginalHeader GetOriginalHeader(const String& variable)
   {
      return (OriginalHeader)FindStringInArray(ms_templateOriginalVars,
                                               OriginalHeader_Max, variable);
   }

   // ctor takes the sink (we need it to implement some pseudo macros such as
   // "$CURSOR") and also a pointer to message for things like $QUOTE - this
   // may (and in fact should) be NULL for new messages, in this case using the
   // macros which require it will result in an error.
   //
   // And we also need the compose view to expand the macros in the "message"
   // category.
   VarExpander(ExpansionSink& sink,
               wxComposeView& cv,
               Profile *profile = NULL,
               Message *msg = NULL)
      : m_sink(sink), m_cv(cv)
   {
      m_profile = profile ? profile : mApplication->GetProfile();
      m_profile->IncRef();
      m_msg = msg;
      SafeIncRef(m_msg);
   }

   virtual ~VarExpander()
   {
      m_profile->DecRef();
      SafeDecRef(m_msg);
   }

   // implement base class pure virtual function
   virtual bool Expand(const String& category,
                       const String& name,
                       const wxArrayString& arguments,
                       String *value) const;

protected:
   // read the file into the string, return TRUE on success
   static bool SlurpFile(const String& filename, String *value);

   // try to find the filename the user wants if only the name was specified
   // (look in the standard locations...)
   static String GetAbsFilename(const String& name);

   // Expand determines the category and then calls one of these functions
   bool ExpandMisc(const String& name, String *value) const;
   bool ExpandFile(const String& name,
                   const wxArrayString& arguments,
                   String *value) const;
   bool ExpandAttach(const String& name,
                     const wxArrayString& arguments,
                     String *value) const;
   bool ExpandCommand(const String& name, String *value) const;
#ifdef USE_PYTHON
   bool ExpandPython(const String& name, String *value) const;
#endif // USE_PYTHON
   bool ExpandMessage(const String& name, String *value) const;
   bool ExpandOriginal(const String& name, String *value) const;

private:
   // helper used by GetCategory and GetVariable
   static int FindStringInArray(const char *strs[], int max, const String& s);

   // the sink we use when expanding pseudo variables
   ExpansionSink& m_sink;

   // the compose view is used for expansion of the variables in "message"
   // category
   wxComposeView& m_cv;

   // the message used for expansion of variables pertaining to the original
   // message (may be NULL for new messages)
   Message *m_msg;

   // the profile to use for everything (global one by default)
   Profile *m_profile;

   // this array contains the list of all categories
   static const char *ms_templateVarCategories[Category_Max];

   // this array contains the list of all variables without category
   static const char *ms_templateVarNames[Var_Max];

   // this array contains the variable names from "message" category
   static const char *ms_templateMessageVars[MessageHeader_Max];

   // this array contains all the variables in the "original" category which
   // map to the headers of the original message (there are other variables in
   // this category as well)
   static const char *ms_templateOriginalVars[OriginalHeader_Max];
};

// ----------------------------------------------------------------------------
// global data: the definitions of the popum menu for the template editing
// dialog.
// ----------------------------------------------------------------------------

// NB: These menus should be kept in sync (if possible) with the variable names

// the misc submenu
static TemplatePopupMenuItem gs_popupSubmenuMisc[] =
{
   TemplatePopupMenuItem(_("Put &cursor here"), "$cursor"),
   TemplatePopupMenuItem(),
   TemplatePopupMenuItem(_("Insert current &date"), "$date"),
   TemplatePopupMenuItem(),
   TemplatePopupMenuItem(_("Insert &quoted text"), "$quote"),
   TemplatePopupMenuItem(_("&Attach original text"), "$quote822"),
};

// the file insert/attach sub menu
static TemplatePopupMenuItem gs_popupSubmenuFile[] =
{
   TemplatePopupMenuItem(_("&Insert file..."), "${file:%s}", TRUE),
   TemplatePopupMenuItem(_("Insert &any file..."), "${file:%s?ask", TRUE),
   TemplatePopupMenuItem(_("Insert &quoted file..."), "${file:%s?quote}", TRUE),
   TemplatePopupMenuItem(_("&Attach file..."), "${attach:%s}", TRUE),
};

// the message submenu
static TemplatePopupMenuItem gs_popupSubmenuMessage[] =
{
   TemplatePopupMenuItem(_("&To"), "${message:to}"),
   TemplatePopupMenuItem(_("&First name"), "${message:firstname}"),
   TemplatePopupMenuItem(_("&Last name"), "${message:lastname}"),
   TemplatePopupMenuItem(_("&Subject"), "${message:subject}"),
   TemplatePopupMenuItem(_("&CC"), "${message:cc}"),
   TemplatePopupMenuItem(_("&BCC"), "${message:bcc}"),
};

// the original message submenu
static TemplatePopupMenuItem gs_popupSubmenuOriginal[] =
{
   TemplatePopupMenuItem(_("&Date"), "${original:date}"),
   TemplatePopupMenuItem(_("&From"), "${original:from}"),
   TemplatePopupMenuItem(_("&Subject"), "${original:subject}"),
   TemplatePopupMenuItem(_("Full &name"), "${original:fullname}"),
   TemplatePopupMenuItem(_("F&irst name"), "${original:firstname}"),
   TemplatePopupMenuItem(_("&Last name"), "${original:lastname}"),
   TemplatePopupMenuItem(_("&To"), "${original:to}"),
   TemplatePopupMenuItem(_("&Reply to"), "${original:replyto}"),
   TemplatePopupMenuItem(_("&Newsgroups"), "${original:newsgroups}"),
};

// the whole menu
static TemplatePopupMenuItem gs_popupMenu[] =
{
   TemplatePopupMenuItem(_("&Miscellaneous"),
                         gs_popupSubmenuMisc,
                         WXSIZEOF(gs_popupSubmenuMisc)),
   TemplatePopupMenuItem(_("Message &headers"),
                         gs_popupSubmenuMessage,
                         WXSIZEOF(gs_popupSubmenuMessage)),
   TemplatePopupMenuItem(_("&Original message"),
                         gs_popupSubmenuOriginal,
                         WXSIZEOF(gs_popupSubmenuOriginal)),
   TemplatePopupMenuItem(_("Insert or attach a &file"),
                         gs_popupSubmenuFile,
                         WXSIZEOF(gs_popupSubmenuFile)),
   TemplatePopupMenuItem(),
   TemplatePopupMenuItem(_("&Execute command..."), "${cmd:%s}", FALSE),
};

const TemplatePopupMenuItem& g_ComposeViewTemplatePopupMenu =
   TemplatePopupMenuItem("", gs_popupMenu, WXSIZEOF(gs_popupMenu));

// ----------------------------------------------------------------------------
// event tables &c
// ----------------------------------------------------------------------------

IMPLEMENT_CLASS(wxComposeView, wxMFrame)

BEGIN_EVENT_TABLE(wxComposeView, wxMFrame)
   // process termination notification
   EVT_END_PROCESS(HelperProcess_Editor, wxComposeView::OnExtEditorTerm)

   // button notifications
   EVT_BUTTON(IDB_EXPAND, wxComposeView::OnExpand)

   // identity combo notification
   EVT_CHOICE(IDC_IDENT_COMBO, wxComposeView::OnIdentChange)
END_EVENT_TABLE()


BEGIN_EVENT_TABLE(wxAddressTextCtrl, wxTextCtrl)
   EVT_SET_FOCUS(wxAddressTextCtrl::OnFocusSet)
   EVT_CHAR(wxAddressTextCtrl::OnChar)
   EVT_TEXT_ENTER(-1, wxAddressTextCtrl::OnEnter)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(wxComposerLayoutWindow, wxLayoutWindow)
   EVT_SET_FOCUS(wxComposerLayoutWindow::OnFocus)
END_EVENT_TABLE()

// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// MimeContent
// ----------------------------------------------------------------------------

void MimeContent::SetMimeType(const String& mimeType)
{
   m_MimeType = mimeType;

   // determin the numeric type
   String category = mimeType.Before('/');
   if ( category.CmpNoCase("VIDEO") == 0 )
      m_NumericMimeType = Message::MSG_TYPEVIDEO;
   else if ( category.CmpNoCase("AUDIO") == 0 )
      m_NumericMimeType = Message::MSG_TYPEAUDIO;
   else if ( category.CmpNoCase("IMAGE") == 0 )
      m_NumericMimeType = Message::MSG_TYPEIMAGE;
   else if ( category.CmpNoCase("TEXT") == 0 )
      m_NumericMimeType = Message::MSG_TYPETEXT;
   else if ( category.CmpNoCase("MESSAGE") == 0 )
      m_NumericMimeType = Message::MSG_TYPEMESSAGE;
   else if ( category.CmpNoCase("APPLICATION") == 0 )
      m_NumericMimeType = Message::MSG_TYPEAPPLICATION;
   else if ( category.CmpNoCase("MODEL") == 0 )
      m_NumericMimeType = Message::MSG_TYPEMODEL;
   else
      m_NumericMimeType = Message::MSG_TYPEOTHER;
}

void MimeContent::SetData(char *data, size_t length, const char *filename)
{
   ASSERT( data != NULL );

   m_Data = data;
   m_Length = length;
   m_Type = MIMECONTENT_DATA;
   if(filename)
      m_FileName = filename;
}

void MimeContent::SetFile(const String& filename)
{
   ASSERT( !filename.IsEmpty() );

   m_FileName = filename;
   m_Type = MIMECONTENT_FILE;
}

// ----------------------------------------------------------------------------
// wxAddressTextCtrl
// ----------------------------------------------------------------------------

// pass to the next control when <Enter> is pressed
void wxAddressTextCtrl::OnEnter(wxCommandEvent& /* event */)
{
    wxNavigationKeyEvent event;
    event.SetDirection(TRUE);       // forward
    event.SetWindowChange(FALSE);   // control change
    event.SetEventObject(this);

    GetParent()->GetEventHandler()->ProcessEvent(event);
}

// mark this ctrl as being the last active - so the [Expand] btn will expand us
void wxAddressTextCtrl::OnFocusSet(wxFocusEvent& event)
{
   m_composeView->SetLastAddressEntry(m_id);

   event.Skip();
}

// expand the address when <TAB> is pressed
void wxAddressTextCtrl::OnChar(wxKeyEvent& event)
{
   ASSERT( event.GetEventObject() == this ); // how can we get anything else?

   m_composeView->SetLastAddressEntry(m_id);

   // we're only interested in TABs and only it's not a second TAB in a row
   if ( event.KeyCode() == WXK_TAB )
   {
      if ( IsModified() &&
           !event.ControlDown() && !event.ShiftDown() && !event.AltDown() )
      {
         // mark control as being "not modified" - if the user presses TAB
         // the second time go to the next window immediately after having
         // expanded the entry
         DiscardEdits();

         // DoExpand() normally checks if the text is empty and gives an error
         // message if it's true - because it doesn't make sense to press the
         // [Expand] button then. This behaviour is not desirable when the user
         // presses TAB, however, so check it here and just treat TAB normally
         // if the text is empty instead.
         if ( !GetValue().IsEmpty() )
         {
            DoExpand();

            // don't call event.Skip()
            return;
         }
         //else: text is empty, treat the TAB normally
      }
      //else: nothing because we're not interested in Ctrl-TAB, Shift-TAB &c -
      //      and also in the TABs if the last one was already a TAB
   }

   // let the text control process it normally: if it's a TAB this will make
   // the focus go to the next window
   event.Skip();
}

bool wxAddressTextCtrl::DoExpand()
{
   // try to expand the last component
   String text = GetValue();
   text.Trim(FALSE); // trim spaces from both sides
   text.Trim(TRUE);

   // check for the lone '"' simplifies the code for finding the starting
   // position below: it should be done here, otherwise the following loop
   // will crash!
   if ( text.IsEmpty() || text == '"' )
   {
      // don't do anything
      wxLogStatus(m_composeView,
                  _("Nothing to expand - please enter something."));

      return FALSE;
   }

   // find the starting position of the last address in the address list
   size_t nLastAddr;
   bool quoted = text.Last() == '"';
   if ( quoted )
   {
      // just find the matching quote (not escaped)
      const char *pStart = text.c_str();
      const char *p;
      for ( p = pStart + text.length() - 2; p >= pStart; p-- )
      {
         if ( *p == '"' )
         {
            // check that it's not escaped
            if ( (p == pStart) || (*(p - 1) != '\\') )
            {
               // ok, found it!
               break;
            }
         }
      }

      nLastAddr = p - pStart;
   }
   else
   {
      // search back until the last address separator
      for ( nLastAddr = text.length() - 1; nLastAddr > 0; nLastAddr-- )
      {
         char c = text[nLastAddr];
         if ( (c == ',') || (c == ';') )
            break;
      }

      if ( nLastAddr > 0 )
      {
         // move beyond the ',' or ';' which stopped the scan
         nLastAddr++;
      }

      // the address will probably never start with spaces but if it does, it
      // will be enough to just take it into quotes
      while ( isspace(text[nLastAddr]) )
      {
         nLastAddr++;
      }
   }

   wxArrayString expansions;
   if ( AdbExpand(expansions, text.c_str() + nLastAddr, m_lookupMode, m_composeView) )
   {
      // find the end of the previous address
      size_t nPrevAddrEnd;
      if ( nLastAddr > 0 )
      {
         // undo "++" above
         nLastAddr--;
      }

      for ( nPrevAddrEnd = nLastAddr; nPrevAddrEnd > 0; nPrevAddrEnd-- )
      {
         char c = text[nPrevAddrEnd];
         if ( !isspace(c) && (c != ',') && (c != ';') )
         {
            // this character is a part of previous string, leave it there
            nPrevAddrEnd++;

            break;
         }
      }

      // take what was there before...
      wxString newText(text, nPrevAddrEnd);  // first nPrevAddrEnd chars
      if ( !newText.IsEmpty() )
      {
         // there was something before, add separator
         newText += CANONIC_ADDRESS_SEPARATOR;
      }

      // ... and and the replacement string(s)
      size_t nExpCount = expansions.GetCount();
      for ( size_t nExp = 0; nExp < nExpCount; nExp++ )
      {
         if ( nExp > 0 )
            newText += CANONIC_ADDRESS_SEPARATOR;

         wxString address(expansions[nExp]);

         // sometimes we must quote the address
         bool doQuote = strpbrk(address, ",;\"") != (const char *)NULL;
         if ( doQuote )
         {
            newText += '"';

            // escape all quotes
            address.Replace("\"", "\\\"");
         }

         newText += address;

         if ( doQuote )
         {
            newText += '"';
         }
      }

      SetValue(newText);
      SetInsertionPointEnd();
   }

   return TRUE;
}

// ----------------------------------------------------------------------------
// wxComposeView inline functions
// ----------------------------------------------------------------------------

void wxComposeView::EnableEditing(bool enable)
{
   // indicate the current state in the status bar
   SetStatusText(enable ? "" : _("RO"), 1);
   m_LayoutWindow->SetEditable(enable);
}

// ----------------------------------------------------------------------------
// wxComposeView creation
// ----------------------------------------------------------------------------

/** Constructor for posting news.
    @param parentProfile parent profile
    @param hide if true, do not show frame
    @return pointer to the new compose view
*/
wxComposeView *
wxComposeView::CreateNewArticle(const MailFolder::Params& params,
                                Profile *parentProfile,
                                bool hide)
{
   wxWindow *parent = mApplication->TopLevelFrame();
   wxComposeView *cv = new wxComposeView("ComposeViewNews", parent);
   cv->m_mode = Mode_News;
   cv->m_kind = Message_New;
   cv->m_template = params.templ;
   cv->SetTitle(COMPOSER_TITLE);
   cv->Create(parent, parentProfile);

   return cv;
}

/** Constructor for sending mail.
    @param parentProfile parent profile
    @param hide if true, do not show frame
    @return pointer to the new compose view
*/
wxComposeView *
wxComposeView::CreateNewMessage(const MailFolder::Params& params,
                                Profile *parentProfile,
                                bool hide)
{
   wxWindow *parent = mApplication->TopLevelFrame();
   wxComposeView *cv = new wxComposeView("ComposeViewMail", parent);
   cv->m_mode = Mode_Mail;
   cv->m_kind = Message_New;
   cv->m_template = params.templ;
   cv->SetTitle(COMPOSER_TITLE);
   cv->Create(parent,parentProfile);

   return cv;
}

wxComposeView *
wxComposeView::CreateReplyMessage(const MailFolder::Params& params,
                                  Profile *parentProfile,
                                  Message *original,
                                  bool hide)
{
   // TODO set the encoding of the original message here for this message

   wxComposeView *cv = CreateNewMessage(parentProfile, hide);

   cv->m_kind = Message_Reply;
   cv->m_template = params.templ;

   cv->m_OriginalMessage = original;
   SafeIncRef(cv->m_OriginalMessage);

   return cv;
}

wxComposeView *
wxComposeView::CreateFwdMessage(const MailFolder::Params& params,
                                Profile *parentProfile,
                                bool hide)
{
   // TODO set the encoding of the original message here for this message

   wxComposeView *cv = CreateNewMessage(parentProfile, hide);

   cv->m_kind = Message_Forward;
   cv->m_template = params.templ;

   return cv;
}

void
wxComposeView::InitText(Message *msg)
{
   if ( m_kind == Message_New )
   {
      // writing a new message/article: wait until the headers are filled
      // before evacuating the template
      if ( !m_template )
      {
         m_template = GetMessageTemplate(m_mode == Mode_Mail
                                          ? MessageTemplate_NewMessage
                                          : MessageTemplate_NewArticle,
                                         m_Profile);
      }

      if ( !!m_template )
      {
         InsertText(_("--- Please choose evaluate template from ---\n"
                      "--- the menu after filling the headers.  ---"));
         ResetDirty();
      }
      else
      {
         // this will only insert the sig
         DoInitText(NULL);
      }
   }
   else
   {
      // replying or forwarding - evaluate the template right now
      CHECK_RET( msg, "no message in InitText" );

      DoInitText(msg);
   }
}

void
wxComposeView::DoInitText(Message *msg)
{
   // first append signature - everything else will be inserted before it
   if( READ_CONFIG(m_Profile, MP_COMPOSE_USE_SIGNATURE) )
   {
      wxTextFile fileSig;

      bool hasSign = false;
      while ( !hasSign )
      {
         String strSignFile = READ_CONFIG(m_Profile, MP_COMPOSE_SIGNATURE);
         if ( !strSignFile.IsEmpty() )
            hasSign = fileSig.Open(strSignFile);

         if ( !hasSign )
         {
            // no signature at all or sig file not found, propose to choose or
            // change it now
            wxString msg;
            if ( strSignFile.IsEmpty() )
            {
               msg = _("You haven't configured your signature file.");
            }
            else
            {
               // to show message from wxTextFile::Open()
               wxLog *log = wxLog::GetActiveTarget();
               if ( log )
                  log->Flush();

               msg.Printf(_("Signature file '%s' couldn't be opened."),
                          strSignFile.c_str());
            }

            msg += _("\n\nWould you like to choose your signature "
                     "right now (otherwise no signature will be used)?");
            if ( MDialog_YesNoDialog(msg, this, MDIALOG_YESNOTITLE,
                                     true, "AskForSig") )
            {
               strSignFile = wxPFileSelector("sig",
                                             _("Choose signature file"),
                                             NULL, ".signature", NULL,
                                             _(wxALL_FILES),
                                             0, this);
            }
            else
            {
               // user doesn't want to use signature file
               break;
            }

            if ( strSignFile.IsEmpty() )
            {
               // user canceled "choose signature" dialog
               break;
            }

            m_Profile->writeEntry(MP_COMPOSE_SIGNATURE, strSignFile);
         }
      }

      if ( hasSign )
      {
         wxLayoutList *layoutList = m_LayoutWindow->GetLayoutList();

         // insert separator optionally
         if( READ_CONFIG(m_Profile, MP_COMPOSE_USE_SIGNATURE_SEPARATOR) ) {
            layoutList->LineBreak();
            layoutList->Insert("--");
         }

         // read the whole file
         size_t nLineCount = fileSig.GetLineCount();
         for ( size_t nLine = 0; nLine < nLineCount; nLine++ ) {
            layoutList->LineBreak();
            layoutList->Insert(fileSig[nLine]);
         }

         // let's respect the netiquette
         static const size_t nMaxSigLines = 4;
         if ( nLineCount > nMaxSigLines )
         {
            wxString msg;
            msg.Printf(_("Your signature is %stoo long: it should "
                        "not be more than %d lines."),
                       nLineCount > 10 ? _("way ") : "", nMaxSigLines);
            MDialog_Message(msg, m_LayoutWindow,
                            _("Signature is too long"),
                            GetPersMsgBoxName(M_MSGBOX_SIGNATURE_LENGTH));

         }

         layoutList->MoveCursorTo(wxPoint(0,0));
      }
      else
      {
         // don't ask the next time
         m_Profile->writeEntry(MP_COMPOSE_USE_SIGNATURE, false);
      }
   }

   // now deal with templates: first decide what kind of template do we need
   MessageTemplateKind kind;
   switch ( m_kind )
   {
      case Message_Reply:
         kind = m_mode == Mode_Mail ? MessageTemplate_Reply
                                    : MessageTemplate_Followup;
         break;

      case Message_Forward:
         ASSERT_MSG( m_mode == Mode_Mail, "can't forward article in news" );

         kind = MessageTemplate_Forward;
         break;

      default:
         FAIL_MSG("unknown message kind");
         // still fall through

      case Message_New:
         kind = m_mode == Mode_Mail ? MessageTemplate_NewMessage
                                    : MessageTemplate_NewArticle;
         break;
   }

   // loop until the template is correct or the user abandons the idea to fix
   // it
   bool templateChanged;
   do
   {
      // get the template value
      String templateValue = !m_template ? GetMessageTemplate(kind, m_Profile)
                                         : m_template;

      if ( !templateValue )
      {
         // if there is no template just don't do anything
         break;
      }

      // we will only run this loop once unless there are erros in the template
      // and the user changed it
      templateChanged = FALSE;

      // do parse the template
      ExpansionSink sink;
      VarExpander expander(sink, *this, m_Profile, msg);
      MessageTemplateParser parser(templateValue, _("template"), &expander);
      if ( parser.Parse(sink) )
      {
         sink.InsertTextInto(*this);
      }
      else
      {
         // first show any errors which the call to Parse() above could
         // generate
         wxLog::FlushActive();

         if ( MDialog_YesNoDialog(_("There were errors in the template. Would "
                                    "you like to edit it now?"),
                                    this,
                                    MDIALOG_YESNOTITLE,
                                    true,
                                    "FixTemplate") )
         {
            // depending on whether we had the template from the beginning or
            // whether we use the standard one, we want to propose different
            // fixes

            if ( !m_template )
            {
               // invoke the template configuration dialog and if something
               // changed...
               if ( ConfigureTemplates(m_Profile, this) )
               {
                  // ...restart the loop
                  templateChanged = TRUE;
               }
            }
            else
            {
               // invoke the dialog for all templates
               String templNew = ChooseTemplateFor(kind, this);
               if ( m_template != templNew )
               {
                  templateChanged = TRUE;

                  m_template = templNew;
               }
            }
         }
      }
   } while ( templateChanged );

   // finally, attach a vCard if requested
   if ( READ_CONFIG(m_Profile, MP_USEVCARD) )
   {
      // read the vCard from the file specified in config
      wxString vcard,
               filename = READ_CONFIG(m_Profile, MP_VCARD);

      bool hasCard = false;
      while ( !hasCard )
      {
         wxString msg;
         if ( !filename )
         {
            msg = _("You haven't configured your vCard file.");
         }
         else // have file name
         {
            wxFFile file(filename);
            if ( file.IsOpened() && file.ReadAll(&vcard) )
            {
               hasCard = true;
            }
            else
            {
               wxLog::FlushActive();

               msg.Printf(_("Couldn't read vCard file '%s'."),
                          filename.c_str());
            }
         }

         if ( !hasCard )
         {
            msg += _("\n\nWould you like to choose your vCard now?");
            if ( MDialog_YesNoDialog(msg, this, MDIALOG_YESNOTITLE,
                                     true, "AskForVCard") )
            {
               wxString pattern;
               pattern << _("vCard files (*.vcf)|*.vcf") << '|' << wxALL_FILES;
               filename = wxPFileSelector("vcard",
                                          _("Choose vCard file"),
                                          NULL, "vcard.vcf", NULL,
                                          pattern,
                                          0, this);
            }
            else
            {
               // user doesn't want to use signature file
               break;
            }

            if ( !filename )
            {
               // the file selection dialog was cancelled
               break;
            }
         }
      }

      // now do attach it
      if ( hasCard )
      {
         InsertData(strutil_strdup(vcard), vcard.length(),
                    "text/x-vcard", filename);
      }
   }

   m_LayoutWindow->Refresh();
}

void
wxComposeView::Create(wxWindow * WXUNUSED(parent),
                      Profile *parentProfile,
                      bool hide)
{
   CHECK_RET( !m_initialised, "wxComposeView created twice" );

   m_LayoutWindow = NULL;
   nextFileID = 0;
   m_encoding = wxFONTENCODING_SYSTEM;

   if(!parentProfile)
      parentProfile = mApplication->GetProfile();
   m_Profile = parentProfile;
   m_Profile->IncRef();

   // sometimes this profile had been created before the identity changed:
   // make sure we use the current identity for the new message composition
   m_Profile->SetIdentity(READ_APPCONFIG(MP_CURRENT_IDENTITY));

   // build menu
   // ----------
   AddFileMenu();
   AddEditMenu();
   WXADD_MENU(GetMenuBar(), COMPOSE, _("&Compose"));
   AddLanguageMenu();
   AddHelpMenu();

   // FIXME: provide some visual feedback for them, like
   // enabling/disabling them. Not used yet.
   m_MItemCut = GetMenuBar()->FindItem(WXMENU_EDIT_CUT);
   m_MItemCopy = GetMenuBar()->FindItem(WXMENU_EDIT_COPY);
   m_MItemPaste = GetMenuBar()->FindItem(WXMENU_EDIT_CUT);

   AddToolbarButtons(CreateToolBar(), WXFRAME_COMPOSE);

   CreateStatusBar(2);
   static const int s_widths[] = { -1, 90 };
   SetStatusWidths(WXSIZEOF(s_widths), s_widths);

   // check if we can schedule messages:
   MModule *module = MModule::GetProvider(MMODULE_INTERFACE_CALENDAR);
   if(module == NULL)
   {
      // if menu is !NULL, it will be filled with wxMenu this item belongs to
      wxMenu *menu = NULL;
      GetMenuBar()->FindItem(WXMENU_COMPOSE_SEND_LATER, &menu);
      ASSERT(menu);
      menu->Delete(WXMENU_COMPOSE_SEND_LATER);
   }
   else
      module->DecRef(); // we don't need it yet

   // create the child controls
   // -------------------------

   // NB: order of item creation is important for constraints algorithm:
   //     create first the controls whose constraints can be fixed first!
   wxLayoutConstraints *c;
   SetAutoLayout(TRUE);

   // Panel itself (fills all the frame client area)
   m_panel = new wxPanel(this, -1);
   c = new wxLayoutConstraints;
   c->top.SameAs(this, wxTop);
   c->left.SameAs(this, wxLeft);
   c->width.SameAs(this, wxWidth);
   c->height.SameAs(this, wxHeight);
   m_panel->SetConstraints(c);
   m_panel->SetAutoLayout(TRUE);

   // box which contains all the header fields
   wxStaticBox *box = new wxStaticBox(m_panel, -1, "");
   c = new wxLayoutConstraints;
   c->left.SameAs(m_panel, wxLeft, LAYOUT_MARGIN);
   c->right.SameAs(m_panel, wxRight, LAYOUT_MARGIN);
   c->top.SameAs(m_panel, wxTop);
   box->SetConstraints(c);

   // layout the labels and text fields: label at the left of the field
   bool bDoShow[Field_Max];

   // "From" is only shown to advanced users
   bDoShow[Field_From] = READ_CONFIG(m_Profile, MP_USERLEVEL) >= M_USERLEVEL_ADVANCED;

   // To and subject always there
   bDoShow[Field_To] =
   bDoShow[Field_Subject] = TRUE;
   if( m_mode == Mode_Mail )
   {
      bDoShow[Field_Cc] = READ_CONFIG(m_Profile, MP_SHOWCC) != 0;
      bDoShow[Field_Bcc] = READ_CONFIG(m_Profile, MP_SHOWBCC) != 0;
   }
   else
      bDoShow[Field_Cc] = bDoShow[Field_Bcc] = FALSE; // never for news


   // first determine the longest label caption
   const char *aszLabels[Field_Max];

   aszLabels[Field_From]    = _("From:");
   aszLabels[Field_To]      = (m_mode == Mode_Mail) ?  _("To:") : _("Newsgroups:");
   aszLabels[Field_Subject] = _("Subject:");
   aszLabels[Field_Cc]      = _("CC:");
   aszLabels[Field_Bcc]     = _("BCC:");

   // don't forget to update the labels when you add a new field
   ASSERT( WXSIZEOF(aszLabels) == Field_Max );

   wxClientDC dc(this);
   long width, widthMax = 0;
   size_t n;
   for ( n = 0; n < Field_Max; n++ )
   {
      if ( !bDoShow[n] )
         continue;

      dc.GetTextExtent(aszLabels[n], &width, NULL);
      if ( width > widthMax )
         widthMax = width;
   }
   widthMax += 2*LAYOUT_MARGIN;

   wxStaticText *label;
   wxWindow  *last = NULL;
   for ( n = 0; n < Field_Max; n++ )
   {
      if ( bDoShow[n] )
      {
         if ( IsAddressField(n) ) {
            m_txtFields[n] = new wxAddressTextCtrl(this, (AddressField)n,
                                                   m_panel);
         }
         else {
            // fields which don't contain addresses don't need completition
            m_txtFields[n] = new wxTextCtrl(m_panel, -1, "");
         }

         // text entry goes from right border of the label to the right border
         // of the box except for the first one where we must also leave space
         // for the button
         c = new wxLayoutConstraints;
         c->left.Absolute(widthMax + LAYOUT_MARGIN);
         if ( last )
            c->top.Below(last, LAYOUT_MARGIN);
         else
            c->top.SameAs(box, wxTop, 2*LAYOUT_MARGIN);

         // for some header fields we create helper controls
         wxControl *ctrlExtra;
         if ( n == Field_From )
         {
            // add the identity combobox
            ctrlExtra = CreateIdentCombo(m_panel);
         }
         else if ( n == Field_To )
         {
            // add the extra [Expand] button
            ctrlExtra = new wxButton(m_panel, IDB_EXPAND, _("Expand"));
#if wxUSE_TOOLTIPS
            ctrlExtra->SetToolTip(_("Expand the address using address books"));
#endif // tooltips
         }
         else
         {
            ctrlExtra = (wxControl *)NULL;
         }

         if ( ctrlExtra )
         {
            wxLayoutConstraints *c2 = new wxLayoutConstraints;
            if ( last )
               c2->top.Below(last, LAYOUT_MARGIN);
            else
               c2->top.SameAs(box, wxTop, 2*LAYOUT_MARGIN);
            c2->right.SameAs(box, wxRight, LAYOUT_MARGIN);
            c2->width.AsIs();
            c2->height.AsIs();
            ctrlExtra->SetConstraints(c2);

            c->right.LeftOf(ctrlExtra, LAYOUT_MARGIN);
         }
         else
         {
            c->right.SameAs(box, wxRight, LAYOUT_MARGIN);
         }

         c->height.AsIs();
         m_txtFields[n]->SetConstraints(c);

         last = m_txtFields[n];

         // label is aligned to the left
         label = new wxStaticText(m_panel, -1, aszLabels[n],
                                  wxDefaultPosition, wxDefaultSize,
                                  wxALIGN_RIGHT);
         c = new wxLayoutConstraints;
         c->left.SameAs(box, wxLeft, LAYOUT_MARGIN);
         c->right.Absolute(widthMax);
         c->centreY.SameAs(last, wxCentreY, 2); // @@ looks better with 2...
         c->height.AsIs();
         label->SetConstraints(c);
      }
      else
      { // not shown
         m_txtFields[n] = NULL;
      }
   }

   // now we can fix the bottom of the box
   box->GetConstraints()->bottom.Below(last, -LAYOUT_MARGIN);

   // compose window
   CreateFTCanvas(); // creates a m_LayoutWindow
   m_LayoutWindow->SetStatusBar(GetStatusBar(),0,1);
   m_LayoutWindow->SetCursorVisibility(1);

   c = new wxLayoutConstraints;
   c->left.SameAs(m_panel, wxLeft, LAYOUT_MARGIN);
   c->right.SameAs(m_panel, wxRight, LAYOUT_MARGIN);
   c->top.Below(box, LAYOUT_MARGIN);
   c->bottom.SameAs(m_panel, wxBottom, LAYOUT_MARGIN);
   m_LayoutWindow->SetConstraints(c);

   // initialize the controls
   // -----------------------

   // set def values for the headers
   SetFrom();
   if(m_txtFields[Field_To])
      m_txtFields[Field_To]->SetValue(READ_CONFIG(m_Profile, MP_COMPOSE_TO));
   if(m_txtFields[Field_Cc])
      m_txtFields[Field_Cc]->SetValue(READ_CONFIG(m_Profile, MP_COMPOSE_CC));
   if(m_txtFields[Field_Bcc])
      m_txtFields[Field_Bcc]->SetValue(READ_CONFIG(m_Profile, MP_COMPOSE_BCC));

   // show the frame
   // --------------
   m_initialised = true;
   if ( !hide ) {
      Show(TRUE);
      m_txtFields[Field_To]->SetFocus();
   }
}

#define NUM_FONTS 7
static int wxFonts[NUM_FONTS] =
{
  wxDEFAULT,
  wxDECORATIVE,
  wxROMAN,
  wxSCRIPT,
  wxSWISS,
  wxMODERN,
  wxTELETYPE
};

void
wxComposeView::CreateFTCanvas(void)
{
   wxASSERT_MSG( m_LayoutWindow == NULL, "creating layout window twice?" );
   m_LayoutWindow = new wxComposerLayoutWindow(this, m_panel);

   GetColourByName(&m_fg, READ_CONFIG(m_Profile, MP_CVIEW_FGCOLOUR), "black");
   GetColourByName(&m_bg, READ_CONFIG(m_Profile, MP_CVIEW_BGCOLOUR), "white");

   m_font = READ_CONFIG(m_Profile,MP_CVIEW_FONT);
   ASSERT( m_font >= 0 && m_font <= NUM_FONTS );

   m_font = wxFonts[m_font];
   m_size = READ_CONFIG(m_Profile,MP_CVIEW_FONT_SIZE);
#ifndef OS_WIN
   m_LayoutWindow->SetFocusFollowMode(
      READ_CONFIG(m_Profile, MP_FOCUS_FOLLOWSMOUSE) != 0);
#endif

   EnableEditing(true);
   DoClear();

   m_LayoutWindow->SetWrapMargin( READ_CONFIG(m_Profile, MP_WRAPMARGIN));
   m_LayoutWindow->SetWordWrap( READ_CONFIG(m_Profile, MP_AUTOMATIC_WORDWRAP) != 0);
}

void wxComposeView::DoClear()
{
   m_LayoutWindow->Clear(m_font, m_size, (int) wxNORMAL, (int) wxNORMAL, 0,
                         &m_fg, &m_bg);

   // set the default encoding if any
   m_encoding = (wxFontEncoding)READ_CONFIG(m_Profile, MP_MSGVIEW_DEFAULT_ENCODING);
   if ( m_encoding != wxFONTENCODING_DEFAULT )
   {
      CheckLanguageInMenu(this, m_encoding);
      m_LayoutWindow->GetLayoutList()->SetFontEncoding(m_encoding);
   }

   ResetDirty();
}

wxComposeView::wxComposeView(const String &iname,
                             wxWindow *parent)
             : wxMFrame(iname,parent)
{
   m_initialised = false;

   m_name = iname;
   m_pidEditor = 0;
   m_procExtEdit = NULL;
   m_fieldLast = Field_Max;
   m_sent = false;
   m_OriginalMessage = NULL;
}

wxComposeView::~wxComposeView()
{
   if ( m_initialised )
   {
      m_Profile->DecRef();
      delete m_LayoutWindow;
   }
   SafeDecRef(m_OriginalMessage);
}

// ----------------------------------------------------------------------------
// wxComposeView callbacks
// ----------------------------------------------------------------------------

// use the newly specified identity
void
wxComposeView::OnIdentChange(wxCommandEvent& event)
{
   wxString ident = event.GetString();
   if ( ident != m_Profile->GetIdentity() )
   {
      if ( event.GetInt() == 0 )
      {
         // first one is always the default identity
         m_Profile->ClearIdentity();
      }
      else
      {
         m_Profile->SetIdentity(ident);
      }

      SetFrom();
   }
}

// expand (using the address books) the value of the last active text zone
void
wxComposeView::OnExpand(wxCommandEvent &WXUNUSED(event))
{
   if ( m_fieldLast == Field_Max )
   {
      // assume "To:" by default
      m_fieldLast = Field_To;
   }

   wxCHECK_RET( IsAddressField(m_fieldLast), "no expansion for this field" );

   (void)((wxAddressTextCtrl *)m_txtFields[m_fieldLast])->DoExpand();
}

bool
wxComposeView::CanClose() const
{
   bool canClose = true;

   // we can't close while the external editor is running (I think it will
   // lead to a nice crash later)
   if ( m_procExtEdit )
   {
      wxLogError(_("Please terminate the external editor (PID %d) before "
                   "closing this window."), m_pidEditor);

      canClose = false;
   }
   else if ( m_LayoutWindow && m_LayoutWindow->IsModified() )
   {
      // ask the user if he wants to save the changes
      canClose = MDialog_YesNoDialog
                 (
                  _("There are unsaved changes, close anyway?"),
                  this, // parent
                  MDIALOG_YESNOTITLE,
                  false, // "yes" not default
                  "UnsavedCloseAnyway"
                 );
   }
   else
   {
      canClose = true;
   }

   return canClose;
}

void
wxComposeView::OnMenuCommand(int id)
{
   switch(id)
   {
   case WXMENU_COMPOSE_INSERTFILE:
      {
         wxArrayString filenames;
         size_t nFiles = wxPFilesSelector(filenames, "MsgInsert",
                                          _("Please choose files to insert."),
                                          NULL, "dead.letter", NULL,
                                          _(wxALL_FILES),
                                          wxOPEN |
                                          wxHIDE_READONLY |
                                          wxFILE_MUST_EXIST,
                                          this);
         for ( size_t n = 0; n < nFiles; n++ )
         {
            InsertFile(filenames[n]);
         }
      }
      break;


   case WXMENU_COMPOSE_SEND:
   case WXMENU_COMPOSE_SEND_LATER:
      if ( IsReadyToSend() )
      {
         if ( Send( (id == WXMENU_COMPOSE_SEND_LATER) ) )
         {
            Close();
         }
      }
      break;

   case WXMENU_COMPOSE_SEND_KEEP_OPEN:
      if ( IsReadyToSend() )
      {
         (void)Send();
      }
      break;

   case WXMENU_COMPOSE_PRINT:
      Print();
      break;

   case WXMENU_COMPOSE_CLEAR:
      DoClear();
      break;

   case WXMENU_COMPOSE_EVAL_TEMPLATE:
      if ( !m_LayoutWindow->IsModified() )
      {
         // remove our prompt
         DoClear();
      }

      DoInitText(NULL);
      break;

   case WXMENU_COMPOSE_LOADTEXT:
      {
         String filename = wxPFileSelector("MsgInsertText",
                                           _("Please choose a file to insert."),
                                           NULL, "dead.letter", NULL,
                                           _(wxALL_FILES),
                                           wxOPEN | wxHIDE_READONLY | wxFILE_MUST_EXIST,
                                           this);

         if ( filename.IsEmpty() )
         {
            wxLogStatus(this, _("Cancelled"));
         }
         else if ( InsertFileAsText(filename) )
         {
            wxLogStatus(this, _("Inserted file '%s'."), filename.c_str());
         }
         else
         {
            wxLogError(_("Failed to insert the text file '%s'."),
                       filename.c_str());
         }
      }
      break;

   case WXMENU_COMPOSE_SAVETEXT:
      {
         String filename = wxPFileSelector("MsgSaveText",
                                           _("Choose file to append message to"),
                                           NULL, "dead.letter", NULL,
                                           _(wxALL_FILES),
                                           wxSAVE,
                                           this);

         if ( filename.IsEmpty() )
         {
            wxLogStatus(this, _("Cancelled"));
         }
         else if ( SaveMsgTextToFile(filename) )
         {
            wxLogStatus(this, _("Message text saved to file '%s'."),
                        filename.c_str());
         }
         else
         {
            wxLogError(_("Failed to save the message."));
         }
      }
      break;

   case WXMENU_COMPOSE_EXTEDIT:
      StartExternalEditor();
      break;

   case WXMENU_COMPOSE_CUSTOM_HEADERS:
      {
         String headerName, headerValue;
         bool storedInProfile;
         if ( ConfigureCustomHeader(m_Profile, this,
                                    &headerName, &headerValue,
                                    &storedInProfile,
                                    m_mode == Mode_News ? CustomHeader_News
                                                        : CustomHeader_Mail) )
         {
            if ( !storedInProfile )
            {
               AddHeaderEntry(headerName, headerValue);
            }
            //else: we will take it from the profile when we will send the msg

            wxLogStatus(this, _("Added custom header '%s' to the message."),
                        headerName.c_str());
         }
         //else: cancelled
      }
      break;

   case WXMENU_HELP_CONTEXT:
      mApplication->Help(
         (m_mode == Mode_Mail)?
         MH_COMPOSE_MAIL : MH_COMPOSE_NEWS, this);
      break;
   case WXMENU_EDIT_PASTE:
      m_LayoutWindow->Paste( WXLO_COPY_FORMAT, FALSE );
      m_LayoutWindow->Refresh();
      break;
   case WXMENU_EDIT_COPY:
      m_LayoutWindow->Copy( WXLO_COPY_FORMAT, FALSE );
      m_LayoutWindow->Refresh();
      break;
   case WXMENU_EDIT_CUT:
      m_LayoutWindow->Cut( WXLO_COPY_FORMAT, FALSE );
      m_LayoutWindow->Refresh();
      break;

   default:
      if ( WXMENU_CONTAINS(LANG, id) && (id != WXMENU_LANG_SET_DEFAULT) )
      {
         m_encoding = GetEncodingFromMenuCommand(id);
         m_LayoutWindow->GetLayoutList()->SetFontEncoding(m_encoding);
         CheckLanguageInMenu(this, m_encoding);
      }
      else
      {
         wxMFrame::OnMenuCommand(id);
      }
   }
}

// ----------------------------------------------------------------------------
// external editor support
// ----------------------------------------------------------------------------

void wxComposeView::OnFirstTimeFocus()
{
   // now we can launch the ext editor if configured to do it
   if ( READ_CONFIG(m_Profile, MP_ALWAYS_USE_EXTERNALEDITOR) )
   {
      if ( !StartExternalEditor() )
      {
         // disable it for the next time
         m_Profile->writeEntry(MP_ALWAYS_USE_EXTERNALEDITOR, 0l);
      }
   }
}

bool wxComposeView::StartExternalEditor()
{
   if ( m_procExtEdit )
   {
      wxLogError(_("External editor is already running (PID %d)"),
                 m_pidEditor);

      // ext editor is running
      return true;
   }

   bool launchedOk = false;

   // if the editor can't be started we ask the user if he wantes to
   // reconfigure it and in this case we retry with the new setting
   bool tryAgain = true;
   while ( !m_procExtEdit && tryAgain )
   {
      tryAgain = false;

      // this entry is supposed to contain '%s' - not to be expanded
      ProfileEnvVarSave envVarDisable(mApplication->GetProfile());

      String extEdit = READ_APPCONFIG(MP_EXTERNALEDITOR);
      if ( !extEdit )
      {
         wxLogStatus(this, _("External editor not configured."));
      }
      else
      {
         // first write the text we already have into a temp file
         MTempFileName tmpFileName;
         if ( !tmpFileName.IsOk() )
         {
            wxLogSysError(_("Cannot create temporary file"));

            break;
         }

         // 'false' means that it's ok to leave the file empty
         if ( !SaveMsgTextToFile(tmpFileName.GetName(), false) )
         {
            wxLogError(_("Failed to pass message to external editor."));

            break;
         }

         // we have a handy function in wxFileType which will replace
         // '%s' with the file name or add the file name at the end if
         // there is no '%s'
         wxFileType::MessageParameters params(tmpFileName.GetName(), "");
         String command = wxFileType::ExpandCommand(extEdit, params);

         // do start the external process
         m_procExtEdit = new wxProcess(this, HelperProcess_Editor);
         m_pidEditor = wxExecute(command, FALSE, m_procExtEdit);

         if ( !m_pidEditor  )
         {
            wxLogError(_("Execution of '%s' failed."), command.c_str());
         }
         else // editor launched
         {
            tmpFileName.Ok();
            m_tmpFileName = tmpFileName.GetName();

            wxLogStatus(this, _("Started external editor (PID %d)"),
                        m_pidEditor);

            // disable editing with the internal editor to avoid
            // interference with the external one
            EnableEditing(false);

            // and reflect it in the title
            wxFrame *frame = GetFrame(this);
            if ( frame )
            {
               wxString title;
               title << COMPOSER_TITLE
                     << _(" (frozen: external editor running)");

               frame->SetTitle(title);
            }
            else
            {
               FAIL_MSG( "composer without frame?" );
            }

            launchedOk = true;
         }
      }

      if ( !m_pidEditor  )
      {
         if ( m_procExtEdit )
         {
            delete m_procExtEdit;
            m_procExtEdit = NULL;
         }

         // either it wasn't configured at all, or the configured editor
         // couldn't be started - propose to change it
         String msg = _("Would you like to change the external "
                        "editor setting now?");
         if ( MDialog_YesNoDialog(msg, this, MDIALOG_YESNOTITLE,
                                  true, "AskForExtEdit") )
         {
            if ( MInputBox(&extEdit,
                           _("Set up external editor"),
                           _("Enter the command name (%s will be "
                             "replaced with the name of the file):"),
                           this,
                           NULL,
                           extEdit) )
            {
               // the ext editor setting is global, don't write it in
               // our profile
               mApplication->GetProfile()->writeEntry(MP_EXTERNALEDITOR,
                                                      extEdit);

               tryAgain = true;
            }
            //else: the dialog was cancelled, don't retry
         }
         //else: user doesn't want to set up ext editor, don't retry
      }
   }

   return launchedOk;
}

void wxComposeView::OnExtEditorTerm(wxProcessEvent& event)
{
   CHECK_RET( event.GetPid() == m_pidEditor , "unknown program terminated" );

   // reenable editing the text in the built-in editor
   EnableEditing(true);

   wxFrame *frame = GetFrame(this);
   if ( frame )
   {
      frame->SetTitle(COMPOSER_TITLE);
   }
   else
   {
      FAIL_MSG( "composer without frame?" );
   }

   bool ok = false;

   // check the return code of the editor process
   if ( event.GetExitCode() != 0 )
   {
      wxLogError(_("External editor terminated with non null exit code."));
   }
   else
   {
      // 'true' means "replace the first text part"
      if ( !InsertFileAsText(m_tmpFileName, true) )
      {
         wxLogError(_("Failed to insert back the text from external editor."));
      }
      else
      {
         if ( remove(m_tmpFileName) != 0 )
         {
            wxLogDebug("Stale temp file '%s' left.", m_tmpFileName.c_str());
         }

         ok = true;
         wxLogStatus(this, _("Inserted text from external editor."));
      }
   }

   if ( !ok )
   {
      wxLogError(_("The text was left in the file '%s'."),
                 m_tmpFileName.c_str());
   }

   m_pidEditor = 0;
   m_tmpFileName.Empty();

   delete m_procExtEdit;
   m_procExtEdit = NULL;
}

// ----------------------------------------------------------------------------
// wxComposeView operations
// ----------------------------------------------------------------------------

void
wxComposeView::InsertData(char *data,
                          size_t length,
                          const char *mimetype,
                          const char *filename)
{
   MimeContent *mc = new MimeContent();

   if(strutil_isempty(mimetype))
      mimetype = "APPLICATION/OCTET-STREAM";


   mc->SetMimeType(mimetype);
   mc->SetData(data, length, filename);

   wxIcon icon =
      mApplication->GetIconManager()->GetIconFromMimeType(mimetype,
                                                          wxString(filename).AfterLast('.'));

   wxLayoutObjectIcon *obj = new wxLayoutObjectIcon(icon);
   obj->SetUserData(mc);
   mc->DecRef();
   m_LayoutWindow->GetLayoutList()->Insert(obj);

   Refresh();
}

void
wxComposeView::InsertFile(const char *fileName, const char *mimetype)
{
   CHECK_RET( !strutil_isempty(fileName), "filename can't be empty" );

   MimeContent
      *mc = new MimeContent();

   // if there is a slash after the dot, it is not extension (otherwise it
   // might be not an extension too, but consider that it is - how can we
   // decide otherwise?)
   String filename(fileName);
   String strExt = filename.AfterLast('.');
   if ( strExt == filename || strchr(strExt, '/') )
      strExt.Empty();

   String strMimeType;
   if ( strutil_isempty(mimetype) )
   {
      strMimeType = GetMimeTypeFromFilename(filename);
   }
   else
   {
      strMimeType = mimetype;
   }

   String msg;
   msg.Printf(_("File '%s' seems to contain data of\n"
                "MIME type '%s'.\n"
                "Is this correct?"),
              filename.c_str(), strMimeType.c_str());
   if ( !MDialog_YesNoDialog( msg, this, _("Content MIME type"),
                              true,
                              GetProfile()->GetName()+"/MimeTypeCorrect") )
   {
      wxString newtype = strMimeType;
      if(MInputBox(&newtype, _("MIME type"),
                   _("Please enter new MIME type:"),
                   this))
         strMimeType = newtype;
   }
   mc->SetMimeType(strMimeType);
   mc->SetFile(filename);

   wxIconManager *iconManager = mApplication->GetIconManager();
   wxIcon icon = iconManager->GetIconFromMimeType(strMimeType, strExt);

   wxLayoutObjectIcon *obj = new wxLayoutObjectIcon(icon);
   obj->SetUserData(mc);
   mc->DecRef();
   m_LayoutWindow->GetLayoutList()->Insert(obj);

   wxLogStatus(this, _("Inserted file '%s' (as '%s')"),
               filename.c_str(), strMimeType.c_str());

   m_LayoutWindow->Refresh();
}

bool
wxComposeView::Send(bool schedule)
{
   Protocol proto;
   switch(m_mode)
   {
      case Mode_Mail:
         proto = READ_CONFIG(m_Profile, MP_USE_SENDMAIL) ? Prot_Sendmail
                                                         : Prot_SMTP;
         break;

      case Mode_News:
         proto = Prot_NNTP;
         break;

      default:
         proto = Prot_Illegal;
   }

   if ( proto == Prot_Illegal )
   {
      FAIL_MSG( "unknown protocol" );

      // we do have to use something though
      proto = Prot_Default;
   }

   // Create the message to be composed
   SendMessage *msg = SendMessage::Create(m_Profile, proto);
   if ( !msg )
   {
      // this shouldn't ever happen, but who knows...
      wxLogError(_("Failed to create the message to send."));

      return false;
   }

   wxBusyCursor bc;
   Disable();

   // compose the body
   wxLayoutObject *lo = NULL;
   MimeContent *mc = NULL;
   wxLayoutExportObject *exp;
   wxLayoutExportStatus status(m_LayoutWindow->GetLayoutList());
   while((exp = wxLayoutExport( &status,
                                   WXLO_EXPORT_AS_TEXT,
                                   WXLO_EXPORT_WITH_CRLF)) != NULL)
   {
      if(exp->type == WXLO_EXPORT_TEXT)
      {
         String* text = exp->content.text;
         msg->AddPart(
                        Message::MSG_TYPETEXT,
                        text->c_str(), text->length(),
                        "PLAIN",
                        "INLINE",   // disposition
                        NULL,       // disposition parameters
                        NULL,       // other parameters
                        m_encoding
                     );
      }
      else
      {
         lo = exp->content.object;
         if(lo->GetType() == WXLO_TYPE_ICON)
         {
            mc = (MimeContent *)lo->GetUserData();

            switch( mc->GetType() )
            {
            case MimeContent::MIMECONTENT_FILE:
            {
               String filename = mc->GetFileName();
               wxFile file;
               if ( file.Open(filename) )
               {
                  size_t size = file.Length();
                  char *buffer = new char[size + 1];
                  if ( file.Read(buffer, size) )
                  {
                     // always NUL terminate it
                     buffer[size] = '\0';

                     MessageParameterList plist, dlist;
                     // some mailers want "FILENAME" in disposition parameters
                     MessageParameter *p = new MessageParameter;
                     p->name = "FILENAME";
                     p->value = wxFileNameFromPath(filename);
                     dlist.push_back(p);
                     // some mailers want "NAME" in parameters:
                     p = new MessageParameter;
                     p->name = "NAME";
                     p->value = wxFileNameFromPath(filename);
                     plist.push_back(p);

                     msg->AddPart
                        (
                           mc->GetMimeCategory(),
                           buffer, size,
                           strutil_after(mc->GetMimeType(),'/'), //subtype
                           "INLINE",
                           &dlist, &plist
                           );
                  }
                  else
                  {
                     wxLogError(_("Cannot read file '%s' included in "
                                  "this message!"), filename.c_str());
                  }

                  delete [] buffer;
               }
               else
               {
                  wxLogError(_("Cannot open file '%s' included in "
                               "this message!"), filename.c_str());
               }
            }
            break;

            case MimeContent::MIMECONTENT_DATA:
            {
               MessageParameterList dlist, plist;
               if(! strutil_isempty(mc->GetFileName()))
               {
                  MessageParameter *p = new MessageParameter;
                  p->name = "FILENAME";
                  p->value = wxFileNameFromPath(mc->GetFileName());
                  dlist.push_back(p);
                  p = new MessageParameter;
                  p->name = "NAME";
                  p->value = wxFileNameFromPath(mc->GetFileName());
                  plist.push_back(p);
               }
               msg->AddPart
                  (
                     mc->GetMimeCategory(),
                     mc->GetData(), mc->GetSize(),
                     strutil_after(mc->GetMimeType(),'/'),  //subtype
                     "INLINE"
                     ,&dlist,&plist
                     );
            }
            break;

            default:
               FAIL_MSG(_("Unknwown part type"));
            }
            mc->DecRef();
         }
      }

      delete exp;
   }

   // setup the headers

   if ( m_encoding != wxFONTENCODING_DEFAULT )
   {
      msg->SetHeaderEncoding(m_encoding);
   }

   // first the standard ones
   msg->SetSubject(GetHeaderValue(Field_Subject));
   switch(m_mode)
   {
      case Mode_Mail:
      {
         // although 'To' field is always present, the others may not be shown
         // (nor created) at all (but now this is checked in GetHeaderValue
         // which will return empty value for them then)
         String to = GetHeaderValue(Field_To);
         String cc = GetHeaderValue(Field_Cc);
         String bcc = GetHeaderValue(Field_Bcc);

         msg->SetAddresses(to, cc, bcc);
      }
      break;

      case Mode_News:
         msg->SetNewsgroups(GetHeaderValue(Field_To));
         break;
   }

   String from = GetHeaderValue(Field_From);
   if ( from && from != m_from )
   {
      msg->SetFrom(from,
                   "" /* don't change the personal name */
                   /* don't set Reply-To yet FIXME?*/);
   }

   // Add additional header lines: first for this time only and then also the
   // headers stored in the profile
   kbStringList::iterator
      i, i2;
   i = m_ExtraHeaderLinesNames.begin();
   i2 = m_ExtraHeaderLinesValues.begin();
   for(; i != m_ExtraHeaderLinesNames.end(); i++, i2++)
      msg->AddHeaderEntry(**i, **i2);

   wxArrayString headerNames, headerValues;
   size_t nHeaders = GetCustomHeaders(m_Profile,
                                      m_mode == Mode_Mail ? CustomHeader_Mail
                                                          : CustomHeader_News,
                                      &headerNames,
                                      &headerValues);
   for ( size_t nHeader = 0; nHeader < nHeaders; nHeader++ )
   {
      msg->AddHeaderEntry(headerNames[nHeader], headerValues[nHeader]);
   }

   // and now do send the message
   bool success;
   if ( schedule )
   {
      MModule_Calendar *calmod =
         (MModule_Calendar *) MModule::GetProvider(MMODULE_INTERFACE_CALENDAR);
      if(calmod)
      {
         success = calmod->ScheduleMessage(msg);
         calmod->DecRef();
      }
      else
      {
         wxLogError(_("Cannot schedule message for sending later because "
                      "the calendar module is not available."));

         success = FALSE;
      }
   }
   else
   {
      success = msg->SendOrQueue();
   }

   delete msg;

   if ( success )
   {
      if ( m_OriginalMessage != NULL )
      {
         // we mark the original message as "answered"
         MailFolder *mf = m_OriginalMessage->GetFolder();
         if(mf)
         {
            mf->SetMessageFlag(m_OriginalMessage->GetUId(),
                               MailFolder::MSG_STAT_ANSWERED, true);
         }
      }

      ResetDirty();
      mApplication->UpdateOutboxStatus();
      wxLogStatus(this, _("Message has been sent."));
   }
   else
   {
      wxLogError(_("The message couldn't be sent."));
   }

   // reenable the window disabled previously
   Enable();

   return success;
}

void
wxComposeView::AddHeaderEntry(const String &entry,
                              const String &ivalue)
{
   String
      *name = new String(entry),
      *value = new String(ivalue);

   m_ExtraHeaderLinesNames.push_back(name);
   m_ExtraHeaderLinesValues.push_back(value);
}

// -----------------------------------------------------------------------------
// simple GUI accessors
// -----------------------------------------------------------------------------

/// sets From field using the current profile
void
wxComposeView::SetFrom()
{
   if ( m_txtFields[Field_From] )
   {
      m_from = miscutil_GetFromAddress(m_Profile);
      m_txtFields[Field_From]->SetValue(m_from);
   }
}

/// sets To field
void
wxComposeView::SetAddresses(const String &to,
                            const String &cc,
                            const String &bcc)
{
   wxTextCtrl *text;

   if ( !!to )
   {
      text = m_txtFields[Field_To];
      if ( text )
         text->SetValue(WXCPTR to.c_str());
   }

   if ( !!cc )
   {
      text = m_txtFields[Field_Cc];
      if ( text )
         text->SetValue(WXCPTR cc.c_str());
   }

   if ( !!bcc )
   {
      text = m_txtFields[Field_Bcc];
      if ( text )
         text->SetValue(WXCPTR bcc.c_str());
   }
}

/// sets Subject field
void
wxComposeView::SetSubject(const String &subj)
{
   m_txtFields[Field_Subject]->SetValue(WXCPTR subj.c_str());
}

void wxComposeView::ResetDirty()
{
   if ( m_LayoutWindow )
      m_LayoutWindow->SetModified(false);
}

void
wxComposeView::MoveCursorTo(int x, int y)
{
   m_LayoutWindow->GetLayoutList()->MoveCursorTo(wxPoint(x, y));
}

/// inserts a text
void
wxComposeView::InsertText(const String &txt)
{
   wxLayoutImportText(m_LayoutWindow->GetLayoutList(), txt);
   m_LayoutWindow->ResizeScrollbars(true);
   m_LayoutWindow->SetDirty();
   m_LayoutWindow->GetLayoutList()->ForceTotalLayout();
}

/// print the message
void
wxComposeView::Print(void)
{
#ifdef OS_WIN
   wxGetApp().SetPrintMode(wxPRINT_WINDOWS);
#else
   bool found;
   wxGetApp().SetPrintMode(wxPRINT_POSTSCRIPT);
   //    set AFM path
   PathFinder pf(mApplication->GetGlobalDir()+"/afm", false);
   pf.AddPaths(READ_CONFIG(m_Profile,MP_AFMPATH), false);
   pf.AddPaths(mApplication->GetLocalDir(), true);
   String afmpath = pf.FindDirFile("Cour.afm", &found);
   if(found)
      wxSetAFMPath(afmpath);
#endif

   wxPrintDialogData pdd(*((wxMApp *)mApplication)->GetPrintData());
   wxPrinter printer(& pdd);
#ifndef OS_WIN
   wxThePrintSetupData->SetAFMPath(afmpath);
#endif
   wxLayoutPrintout printout(m_LayoutWindow->GetLayoutList(),_("Mahogany: Printout"));
   if ( !printer.Print(this, &printout, TRUE)
        && printer.GetLastError() != wxPRINTER_CANCELLED )
      wxMessageBox(_("There was a problem with printing the message:\n"
                     "perhaps your current printer is not set up correctly?"),
                   _("Printing"), wxOK);
   else
      (* ((wxMApp *)mApplication)->GetPrintData())
         = printer.GetPrintDialogData().GetPrintData();
}

// -----------------------------------------------------------------------------
// helper functions
// -----------------------------------------------------------------------------

/// verify that the message is ready to be sent
bool
wxComposeView::IsReadyToSend() const
{
   // verify that the external editor has terminated
   if ( m_procExtEdit )
   {
      // we don't give the user a choice because it will later leave us with
      // an orphan ext editor process and it's easier this way
      wxLogError(_("The external editor is still opened, please close "
                   "it before sending the message."));

      return false;
   }

   // verify that the network is configured
   bool networkSettingsOk = false;
   while ( !networkSettingsOk )
   {
      String host = READ_CONFIG(mApplication->GetProfile(), MP_SMTPHOST);
      if ( host.IsEmpty() )
      {
         if ( MDialog_YesNoDialog(
                  _("The message can not be sent because the network settings "
                    "are either not configured or incorrect. Would you like to "
                    "change them now?"),
                  this,
                  MDIALOG_YESNOTITLE,
                  true /* yes is default */,
                  "ConfigNetFromCompose") )
         {
            ShowOptionsDialog((wxComposeView *)this, OptionsPage_Network);
         }
         else
         {
            wxLogError(_("Cannot send message - network is not configured."));

            return FALSE;
         }
      }
      else
      {
         // TODO any other vital settings to check?
         networkSettingsOk = true;
      }
   }

   // did we forget the recipients?
   if ( !GetHeaderValue(Field_To) )
   {
      wxLogError(_("Please specify at least one recipient in the \"To:\" "
                   "field!"));

      return false;
   }

   // did we forget the subject?
   if ( !GetHeaderValue(Field_Subject) )
   {
      // this one is not strictly speaking mandatory, so just ask the user
      // about it (giving him a chance to get rid of annoying msg box)
      if ( MDialog_YesNoDialog(
               _("The subject of your message is empty, would you like "
                 "to change it?"),
               this,
               MDIALOG_YESNOTITLE,
               true /* yes is default */,
               "SendemptySubject") )
      {
         m_txtFields[Field_Subject]->SetFocus();

         return false;
      }
   }

   // everything is ok
   return true;
}

/// insert a text file at the current cursor position
bool
wxComposeView::InsertFileAsText(const String& filename,
                                bool replaceFirstTextPart)
{
   // read the text from the file
   char *text = NULL;
   off_t lenFile;
   wxFile file(filename);

   bool ok = file.IsOpened();
   if ( ok )
   {
      lenFile = file.Length();
      if ( lenFile == 0 )
      {
         wxLogVerbose(_("File '%s' is empty, no text to insert."),
                      filename.c_str());
         return true;
      }

      text = new char[lenFile + 1];
      text[lenFile] = '\0';

      ok = file.Read(text, lenFile) != wxInvalidOffset;
   }

   if ( !ok )
   {
      wxLogError(_("Cannot insert text file into the message."));

      if ( text )
         delete [] text;

      return false;
   }

   // insert the text in the beginning of the message replacing the old
   // text if asked for this
   if ( replaceFirstTextPart )
   {
      // VZ: I don't know why exactly does this happen but exporting text and
      //     importing it back adds a '\n' at the end, so this is useful as a
      //     quick workaround for this bug - of course, it's not a real solution
      //     (FIXME)
      size_t index = (size_t)lenFile - 1;
      if ( text[index] == '\n' )
      {
         // check for "\r\n" too
         if ( index > 0 && text[index - 1] == '\r' )
         {
            // truncate one char before
            index--;
         }

         text[index] = '\0';
      }

      // this is not as simple as it sounds, because we normally exported all
      // the text which was in the beginning of the message and it's not one
      // text object, but possibly several text objects and line break
      // objects, so now we must delete them and then recreate the new ones...

      wxLayoutList * layoutList = m_LayoutWindow->GetLayoutList();
      wxLayoutList * other_list = new wxLayoutList;
      wxLayoutObject *obj;
      wxLayoutExportStatus status(layoutList);
      wxLayoutExportObject *exp;
      while((exp = wxLayoutExport( &status,
                                      WXLO_EXPORT_AS_OBJECTS)) != NULL)
      {
         // ignore WXLO_EXPORT_EMPTYLINE:
         if(exp->type == WXLO_EXPORT_OBJECT)
         {
            obj = exp->content.object;
            switch(obj->GetType())
            {
            case WXLO_TYPE_TEXT:
               ; //    do nothing
               break;
            case WXLO_TYPE_ICON:
               other_list->Insert(obj->Copy());
               break;
            default:
               ; // cmd    objects get ignored
            }
         }
         delete exp;
      }
      layoutList->Empty();
      //now we move the non-text objects back:
      wxLayoutExportStatus status2(other_list);
      while((exp = wxLayoutExport( &status2,
                                      WXLO_EXPORT_AS_OBJECTS)) != NULL)
         if(exp->type == WXLO_EXPORT_EMPTYLINE)
            layoutList->LineBreak();
         else
            layoutList->Insert(exp->content.object->Copy());
      delete other_list;
   }

   // now insert the new text
   wxLayoutImportText(m_LayoutWindow->GetLayoutList(), text);
   m_LayoutWindow->ResizeScrollbars(true);
   m_LayoutWindow->SetModified();
   m_LayoutWindow->SetDirty();
   m_LayoutWindow->Refresh();
   delete [] text;
   return true;
}

/// save the first text part of the message to the given file
bool
wxComposeView::SaveMsgTextToFile(const String& filename,
                                 bool errorIfNoText) const
{
   // TODO write (and read later...) headers too!

   // write the text part of the message into a file
   wxFile file(filename, wxFile::write_append);
   if ( !file.IsOpened() )
   {
      wxLogError(_("Cannot open file for the message."));

      return false;
   }

   // export first text part of the message

   wxLayoutExportObject *exp;
   wxLayoutExportStatus status(m_LayoutWindow->GetLayoutList());

   while((exp = wxLayoutExport( &status, WXLO_EXPORT_AS_TEXT)) != NULL)
   {
      // non text objects get ignored
      if(exp->type == WXLO_EXPORT_TEXT)
         if ( !file.Write(*exp->content.text) )
         {
            wxLogError(_("Cannot write message to file."));

            return false;
         }
   }

   ((wxComposeView *)this)->ResetDirty(); // const_cast

   return true;
}

// ----------------------------------------------------------------------------
// ExpansionSink - the sink used with wxComposeView
// ----------------------------------------------------------------------------

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(ArrayAttachmentInfo);

bool
ExpansionSink::Output(const String& text)
{
   if ( !m_hasCursorPosition )
   {
      // update the current x and y position
      // FIXME: this supposes that there is no autowrap - is it true?
      int deltaX = 0, deltaY = 0;
      for ( const char *pc = text.c_str(); *pc; pc++ )
      {
         if ( *pc == '\n' )
         {
            deltaX = -m_x;
            deltaY++;
         }
         else
         {
            deltaX++;
         }
      }

      m_x += deltaX;
      m_y += deltaY;
   }
   //else: we don't have to count anything, we already have the cursor position
   //      and this is all we want

   // we always add this text to the last "component" of text (i.e. the one
   // after the last attachment)
   m_text += text;

   return TRUE;
}

void
ExpansionSink::InsertAttachment(void *data,
                                size_t len,
                                const String& mimetype,
                                const String& filename)
{
   if ( !m_hasCursorPosition )
   {
      // an attachment count as one cursor position
      m_x++;
   }

   // create a new attachment info object (NB: it will be deleted by the array
   // automatically because we pass it by pointer and not by reference)
   m_attachments.Add(new AttachmentInfo(data, len, mimetype, filename));

   // the last component of text becomes the text before this attachment
   m_texts.Add(m_text);

   // and the new last component is empty
   m_text.Empty();
}

void
ExpansionSink::InsertTextInto(wxComposeView& cv) const
{
   size_t nCount = m_texts.GetCount();
   ASSERT_MSG( m_attachments.GetCount() == nCount,
               "something is very wrong in template expansion sink" );

   for ( size_t n = 0; n < nCount; n++ )
   {
      cv.InsertText(m_texts[n]);

      AttachmentInfo& attInfo = m_attachments[n];
      cv.InsertData((char *)attInfo.data, attInfo.len,
                    attInfo.mimetype,
                    attInfo.filename);
   }

   cv.InsertText(m_text);

   // position the cursor - if RememberCursorPosition() hadn't been called, it
   // will be put in (0, 0)
   cv.MoveCursorTo(m_x, m_y);

   // as the inserted text comes from the program, not from the user, don't
   // mark the composer contents as dirty
   cv.ResetDirty();
}

// ----------------------------------------------------------------------------
// VarExpander - used by wxComposeView
// ----------------------------------------------------------------------------

const char *VarExpander::ms_templateVarCategories[] =
{
   "",
   "file",
   "attach",
   "cmd",
#ifdef USE_PYTHON
   "python",
#endif // USE_PYTHON
   "message",
   "original",
};

const char *VarExpander::ms_templateVarNames[] =
{
   "date",
   "cursor",
   "to",
   "subject",
   "quote",
   "quote822",
   "sender",
};

const char *VarExpander::ms_templateMessageVars[] =
{
   // the order must be the same as for the text controls in compose view!
   "to",
   "subject",
   "cc",
   "bcc",
   "firstname",
   "lastname",
};

const char *VarExpander::ms_templateOriginalVars[] =
{
   "date",
   "from",
   "subject",
   "fullname",
   "firstname",
   "lastname",
   "to",
   "replyto",
   "newsgroups",
};

int
VarExpander::FindStringInArray(const char *strings[],
                               int max,
                               const String& s)
{
   int n;
   for ( n = 0; n < max; n++ )
   {
      if ( strings[n] == s )
         break;
   }

   return n;
}

bool
VarExpander::SlurpFile(const String& filename, String *value)
{
   // it's important that value is not empty even if we return FALSE because if
   // it's empty when Expand() returns the template parser will log a
   // misleading error message about "unknown variable"
   *value = '?';

   wxFFile file(filename);

   return file.IsOpened() && file.ReadAll(value);
}

String
VarExpander::GetAbsFilename(const String& name)
{
   String filename = wxExpandEnvVars(name);
   if ( !IsAbsPath(filename) )
   {
      Profile *profile = mApplication->GetProfile();
      String path = profile->readEntry(MP_COMPOSETEMPLATEPATH_USER,
                                       mApplication->GetLocalDir());
      if ( !path || path.Last() != '/' )
      {
         path += '/';
      }
      String filename2 = path + filename;

      if ( wxFile::Exists(filename2) )
      {
         // ok, found
         filename = filename2;
      }
      else
      {
         // try the global dir
         String path = profile->readEntry(MP_COMPOSETEMPLATEPATH_GLOBAL,
                                          mApplication->GetGlobalDir());
         if ( !path || path.Last() != '/' )
         {
            path += '/';
         }

         filename.Prepend(path);
      }
   }
   //else: absolute filename given, don't look anywhere else

   return filename;
}

bool
VarExpander::Expand(const String& category,
                    const String& name,
                    const wxArrayString& arguments,
                    String *value) const
{
   value->Empty();

   // comparison is case insensitive
   switch ( GetCategory(category.Lower()) )
   {
      case Category_File:
         return ExpandFile(name, arguments, value);

      case Category_Attach:
         return ExpandAttach(name, arguments, value);

      case Category_Command:
         return ExpandCommand(name, value);

#ifdef USE_PYTHON
      case Category_Python:
         return ExpandPython(name, value);
#endif // USE_PYTHON

      case Category_Message:
         return ExpandMessage(name, value);

      case Category_Original:
         return ExpandOriginal(name, value);

      case Category_Misc:
         return ExpandMisc(name, value);

      default:
         // unknown category
         return FALSE;
   }
}

bool
VarExpander::ExpandMisc(const String& name, String *value) const
{
   // deal with all special cases
   switch ( GetVariable(name.Lower()) )
   {
      case Var_Date:
         {
            time_t ltime;
            (void)time(&ltime);

#ifdef OS_WIN
            // MP_DATE_FMT contains '%' which are being (mis)interpreted as
            // env var expansion characters under Windows
            ProfileEnvVarSave noEnvVars(m_profile);
#endif // OS_WIN

            *value = strutil_ftime(ltime, READ_CONFIG(m_profile, MP_DATE_FMT));
         }
         break;

      case Var_Cursor:
         m_sink.RememberCursorPosition();
         break;

      case Var_To:
         // just the shorthand for "message:to"
         return ExpandMessage("to", value);

      case Var_Subject:
         return ExpandMessage("subject", value);

      case Var_Quote:
         return ExpandOriginal("quote", value);

      case Var_Quote822:
         return ExpandOriginal("quote822", value);

      case Var_Sender:
         ExpandOriginal("from", value);
         return TRUE;

      default:
         // unknown name
         return FALSE;
   }

   return TRUE;
}

bool
VarExpander::ExpandFile(const String& name,
                        const wxArrayString& arguments,
                        String *value) const
{
   // first check if we don't want to ask user
   String filename = GetAbsFilename(name);
   if ( arguments.Index("ask", FALSE /* no case */) != wxNOT_FOUND )
   {
      filename = MDialog_FileRequester(_("Select the file to insert"),
                                       GetFrame(&m_cv),
                                       filename);
   }

   if ( !!filename )
   {
      // insert the contents of a file
      if ( !SlurpFile(filename, value) )
      {
         wxLogError(_("Failed to insert file '%s' into the message."),
                    name.c_str());

         return FALSE;
      }

      // do we want to quote the files contents before inserting?
      if ( arguments.Index("quote", FALSE /* no case */) != wxNOT_FOUND )
      {
         String prefix = READ_CONFIG(m_profile, MP_REPLY_MSGPREFIX);
         String quotedValue;
         quotedValue.Alloc(value->length());

         const char *cptr = value->c_str();
         quotedValue = prefix;
         while ( *cptr )
         {
            if ( *cptr == '\r' )
            {
               cptr++;
               continue;
            }

            quotedValue += *cptr;
            if( *cptr++ == '\n' && *cptr )
            {
               quotedValue += prefix;
            }
         }

         *value = quotedValue;
      }
   }
   //else: no file, nothing to insert

   return TRUE;
}

bool
VarExpander::ExpandAttach(const String& name,
                          const wxArrayString& arguments,
                          String *value) const
{
   String filename = GetAbsFilename(name);
   if ( arguments.Index("ask", FALSE /* no case */) != wxNOT_FOUND )
   {
      filename = MDialog_FileRequester(_("Select the file to attach"),
                                       GetFrame(&m_cv),
                                       filename);
   }

   if ( !!filename )
   {
      if ( !SlurpFile(filename, value) )
      {
         wxLogError(_("Failed to attach file '%s' to the message."),
                    name.c_str());

         return FALSE;
      }

      // guess MIME type from extension
      m_sink.InsertAttachment(strutil_strdup(value->c_str()),
                              value->length(),
                              GetMimeTypeFromFilename(filename),
                              filename);

      // avoid inserting file as text additionally
      value->Empty();
   }
   //else: no file, nothing to attach

   return TRUE;
}

bool
VarExpander::ExpandCommand(const String& name, String *value) const
{
   // execute a command (FIXME get the wxProcess class allowing stdout
   // redirection and use it here)
   MTempFileName temp;

   bool ok = temp.IsOk();
   wxString filename = temp.GetName(), command;

   if ( ok )
      command << name << " > " << filename;

   if ( ok )
      ok = system(command) == 0;

   if ( ok )
      ok = SlurpFile(filename, value);

   if ( !ok )
   {
      wxLogSysError(_("Failed to execute the command '%s'"), name.c_str());

      return FALSE;
   }

   return TRUE;
}

#ifdef USE_PYTHON
bool
VarExpander::ExpandPython(const String& name, String *value) const
{
   // TODO
   return FALSE;
}
#endif // USE_PYTHON

bool
VarExpander::ExpandMessage(const String& name, String *value) const
{
   MessageHeader header = GetMessageHeader(name.Lower());
   if ( header == MessageHeader_Invalid )
   {
      // unknown variable
      return FALSE;
   }

   switch ( header )
   {
      case MessageHeader_FirstName:
      case MessageHeader_LastName:
         {
            wxString to = m_cv.GetHeaderValue(wxComposeView::Field_To);
            if ( header == MessageHeader_FirstName )
               *value = Message::GetFirstNameFromAddress(to);
            else
               *value = Message::GetLastNameFromAddress(to);
         }
         break;

      default:
         CHECK( header <= MessageHeader_LastControl, FALSE,
                "unexpected macro in message category" );

         // the MessageHeader enum values are the same as AddressField ones
         // except for the shift due to absence of "From", so no translation is
         // needed
         *value = m_cv.GetHeaderValue((wxComposeView::AddressField)(header + 1));
   }

   return TRUE;
}

bool
VarExpander::ExpandOriginal(const String& Name, String *value) const
{
   // insert the quoted text of the original message - of course, this
   // only works if we have this original message
   if ( !m_msg )
   {
      // unfortunately we don't have the real name of the variable
      // here
      wxLogWarning(_("The variables using the original message cannot "
                     "be used in this template; variable ignored."));
   }
   else
   {
      String name = Name.Lower();
      switch ( GetOriginalHeader(name) )
      {
         case OriginalHeader_Date:
            *value = m_msg->Date();
            break;

         case OriginalHeader_From:
            *value = m_msg->From();
            break;

         case OriginalHeader_Subject:
            *value = m_msg->Subject();
            break;

         case OriginalHeader_ReplyTo:
            {
               String dummy;
               *value = m_msg->Address(dummy, MAT_REPLYTO);
            }
            break;

         case OriginalHeader_To:
            {
               String dummy;
               *value = m_msg->Address(dummy, MAT_SENDER);
            }
            break;

         case OriginalHeader_PersonalName:
            m_msg->Address(*value, MAT_FROM);
            break;

         case OriginalHeader_FirstName:
            *value = m_msg->GetAddressFirstName(MAT_FROM);
            break;

         case OriginalHeader_LastName:
            *value = m_msg->GetAddressLastName(MAT_FROM);
            break;

         case OriginalHeader_Newsgroups:
            m_msg->GetHeaderLine("Newsgroups", *value);
            break;

         default:
            // it isn't a variable which maps directly onto header, check the
            // others
            if ( name == "text" || name == "quote" )
            {
               // insert the original text (optionally prefixed by reply
               // string)
               String prefix;
               if ( name == "quote" )
               {
                  // prepend the senders initials to the reply prefix (this
                  // will make reply prefix like "VZ>")
                  if ( READ_CONFIG(m_profile, MP_REPLY_MSGPREFIX_FROM_SENDER) )
                  {
                     String name;
                     m_msg->Address(name);
                     wxStringTokenizer tk(name);
                     while ( tk.HasMoreTokens() )
                     {
                        prefix += tk.GetNextToken()[0u];
                     }
                  }

                  // and then the standard reply prefix too
                  prefix += READ_CONFIG(m_profile, MP_REPLY_MSGPREFIX);
               }
               //else: name == text, so no reply prefix at all

               // should we quote the empty lines?
               bool quoteEmpty = READ_CONFIG(m_profile, MP_REPLY_QUOTE_EMPTY) != 0;
               int nParts = m_msg->CountParts();
               for ( int nPart = 0; nPart < nParts; nPart++ )
               {
                  if ( m_msg->GetPartType(nPart) == Message::MSG_TYPETEXT )
                  {
                     String str = m_msg->GetPartContent(nPart);
                     const char *cptr = str.c_str();
                     String str2 = prefix;
                     while ( *cptr )
                     {
                        if ( *cptr == '\r' )
                        {
                           cptr++;
                           continue;
                        }

                        str2 += *cptr;
                        if( *cptr++ == '\n' && *cptr )
                        {
                           if ( quoteEmpty ||
                                (*cptr != '\r' && *cptr != '\n') )
                           {
                              str2 += prefix;
                           }
                        }
                     }

                     *value += str2;
                  }
               }
            }
            else if ( name == "quote822" )
            {
               // insert the original message as RFC822 attachment
               String str;
               m_msg->WriteToString(str);
               m_sink.InsertAttachment(strutil_strdup(str), str.Length(),
                                       "message/rfc822", "");
            }
            else
            {
               return FALSE;
            }
      }
   }

   return TRUE;
}
